/*
  temp_cont1.ino
  
  This sketch was developed by Synapse(Hiroshi Tanigawa) in 2013.
  This Library is originally distributed at "Synapse's history of making gadgets." 
  <http://www3.big.or.jp/~schaft/hardware/>

  This sketch is free software: you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation, either version 2.1 of the License, or
  (at your option) any later version.

  This sketch  is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with MGLCD.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <SPI.h>
#include <mglcd.h>
#include <SoftwareSerial.h>

// 温度センサに電源を供給するためのピン
#define VCC  8
#define GND  9

// 温度センサのチップセレクト信号
#define SLAVE 10

// ヒーターの制御ピン
#define HEATER_PIN A0 

// 圧電ブザー用のピン
#define BUZZER_PIN 5
#define BUZZER_GND 4

// 温度を読んだ際のエラーコード
#define TEMP_NO_ERROR     0
#define TEMP_SHORT_TO_VCC 1
#define TEMP_SHORT_TO_GND 2
#define TEMP_OPEN_CIRCUIT 4

#define TIME_RESOLUTION 100 // 処理時間の最小単位[ms]
#define PWM_RESOLUTION 20 // PWMの分解能

#define TEMP_PREHEAT 150.0 // 予熱温度の中心値
#define WIDTH_PREHEAT 40.0 // 予熱温度の幅
#define TEMP_REFLOW 220.0 // リフロー温度の中心値
#define WIDTH_REFLOW 40.0 // リフロー温度の幅
#define TEMP_COOL 60.0 // 冷却終了温度

#define TIME_PREHEAT1 16000 // 予熱の初期にヒーターを止める時間[ms](オーバーシュート対策)
#define TIME_PREHEAT2 90000 // 予熱時間[ms]
#define TIME_REFLOW1 9000 // リフローの初期にヒータを止める時間[ms](オーバーシュート対策)
#define TIME_REFLOW2 35000 // リフロー時間[ms]

#define TIME_CHECK 120000 // 加熱チェックの時刻[ms]
#define TEMP_CHECK 100.0 // 加熱チェック時の温度の下限

#define RETRY_LIMIT 5 // 温度計測のリトライ回数の上限

#define TEXT_WIDTH 14 // 使用するテキストの桁数
#define LCD_WIDTH 122 // LCDの水平方向のピクセル数
#define LCD_HEIGHT 32 // LCDの垂直方向のピクセル数
#define GRAPH_WIDTH (LCD_WIDTH-TEXT_WIDTH*6) // グラフの水平方向のピクセル数
#define GRAPH_LOW 10.0 // グラフの下限の温度
#define GRAPH_HIGH 270.0 // グラフの上限の温度

SoftwareSerial MySerial(2,3);
mglcd_SoftwareSerial MGLCD(&MySerial,38400);
unsigned long StartTime1; // 加熱スタート時の時刻[ms]
unsigned long StartTime2; // 現在のステートの始まった時刻[ms]
byte TempHist[GRAPH_WIDTH]; // 過去の温度のグラフの長さ
unsigned long NextTick; // 次のループの時刻

// 温度センサの初期化
void InitializeTemperatureSensor()
{
  // 温度センサの電源供給ピンを必要なら初期化する
#ifdef GND
  pinMode(GND,OUTPUT);
  digitalWrite(GND,LOW); 
#endif
#ifdef VCC
  pinMode(VCC,OUTPUT);
  digitalWrite(VCC,HIGH);
#endif

  // 温度センサのチップセレクトピンを初期化する
  pinMode(SLAVE,OUTPUT);
  digitalWrite(SLAVE,HIGH);
  
  // 圧電ブザーのピンを初期化する
  pinMode(BUZZER_GND,OUTPUT);
  digitalWrite(BUZZER_GND,LOW);
  pinMode(BUZZER_PIN,OUTPUT);
  digitalWrite(BUZZER_PIN,LOW);
  
  // SPIインターフェースを初期化する
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  SPI.setClockDivider(SPI_CLOCK_DIV4);
  SPI.setDataMode(SPI_MODE0);
} // InitializeTemperatureSensor

// 温度センサから温度を読む
byte MeasureTemperature(float &temperature)
{
  unsigned TempMp;  // 14-bitの測定点の温度データ
  unsigned TempInt; // 12-bitのIC内部温度のデータ
  
  // 温度データをSPIバス経由で読む  
  digitalWrite(SLAVE, LOW); // チップセレクト信号をイネーブル
  TempMp  = (unsigned int)SPI.transfer(0x00)<<8; // 測定点の温度データの上位バイトを読む
  TempMp  |= (unsigned int)SPI.transfer(0x00);   // 測定店の温度データの下位バイトを読む
  TempInt = SPI.transfer(0x00)<<8; // IC内部温度データの上位バイトを読む
  TempInt |= SPI.transfer(0x00);   // IC内部温度データの下位バイトを読む
  digitalWrite(SLAVE, HIGH); // チップセレクト信号をディゼーブル
  
  // 温度を計算する
  if((TempMp & 0x0001) != 0) { // エラーあり
    byte result=TEMP_NO_ERROR;
    if ((TempInt & 0x0004) !=0) {
      result|=TEMP_SHORT_TO_VCC;
    } // if
    if ((TempInt & 0x0002) !=0) {
      result|=TEMP_SHORT_TO_GND;
    } // if
    if ((TempInt & 0x0001) !=0) {
      result|=TEMP_OPEN_CIRCUIT;
    } // if
    return result;
  } else { // エラー無し
    if((TempMp & 0x8000) == 0){ // 0℃以上 
      temperature = (TempMp>>2) * 0.25;
    } else { // 氷点下
      temperature = (0x4000-(TempMp>>2))*-0.25;
    } // if
    return TEMP_NO_ERROR;
  } // if
} // MeasureTemperature

// 温度を表示する
void PrintTemperature(float temperature, byte result)
{
  MGLCD.Locate(0,0);
  if(result==TEMP_NO_ERROR) {
    MGLCD.print(F("ｵﾝﾄﾞ "));
    MGLCD.print(temperature);
    MGLCD.print("ﾟC");
  } else {
    if(result&TEMP_SHORT_TO_VCC) {
      MGLCD.print(F("VCC ﾆ ｼｮｰﾄ"));
    } else if(result&TEMP_SHORT_TO_GND) {
      MGLCD.print(F("GND ﾆ ｼｮｰﾄ"));
    } else if(result&TEMP_OPEN_CIRCUIT){
      MGLCD.print(F("ﾀﾞﾝｾﾝ ｹﾝｼｭﾂ"));
    } else {
      MGLCD.print(F("ﾌﾒｲﾅ ｴﾗｰ"));
    } // if
  } // if
  MGLCD.ClearRestOfLine();
} // PrintTemperature

// 状態を表示する
void PrintStatus(const char *status) {
  MGLCD.Locate(0,1);
  MGLCD.print(status);
  if(MGLCD.GetX()>0) MGLCD.ClearRestOfLine();
} // Print Status

// 経過時間を表示する
void PrintTime()
{
  unsigned long ElapseTime; // 経過時間
  int hh,mm,ss; // 時分秒
  char str[15]; // 時刻の文字列
  
  ElapseTime=(millis()-StartTime1)/1000; 
  ss=int(ElapseTime%60);
  mm=int((ElapseTime/60)%60);
  hh=int(ElapseTime/3600);
  sprintf(str,"%02d:%02d:%02d",hh,mm,ss);
  MGLCD.Locate(0,2);
  MGLCD.print(str);
  MGLCD.ClearRestOfLine();

  ElapseTime=(millis()-StartTime2)/1000; 
  ss=int(ElapseTime%60);
  mm=int((ElapseTime/60)%60);
  hh=int(ElapseTime/3600);
  sprintf(str,"%02d:%02d:%02d ",hh,mm,ss);
  MGLCD.Locate(0,3);
  MGLCD.print(str);
  MGLCD.ClearRestOfLine();
} // PrintTime

void setup()
{
  InitializeTemperatureSensor(); // 温度センサの初期化

  // ヒーター制御ピンの初期化
  pinMode(HEATER_PIN,OUTPUT);
  digitalWrite(HEATER_PIN,LOW);  

  Serial.begin(9600); // シリアルを9600bpsで初期化
  
  while(MGLCD.Reset()); // 液晶の初期化
  MGLCD.SetCodeMode(MGLCD_CODE_UTF8); // カナ表示を有効に
  MGLCD.SetTextWindow(0,0,TEXT_WIDTH-1,MGLCD.GetLineNum()-1); // テキストウィンドウの設定
  
  // TempHistの初期化(要らんかも)
  for(int i=0; i<GRAPH_WIDTH; i++) {
    TempHist[i]=0;
  } // for i
  
  NextTick=millis()+TIME_RESOLUTION;
} // setup

// PWMのデューティー比を計算する
int CalculateDuty(float CurrentTemp, float TergetTemp, float TargetWidth)
{
  if(CurrentTemp<TergetTemp-TargetWidth/2.0) {
    return PWM_RESOLUTION;
  } else if(CurrentTemp>TergetTemp+TargetWidth/2.0) {
    return 0;
  } else {
    return int(((TergetTemp-CurrentTemp)/TargetWidth+0.5)*PWM_RESOLUTION);
  } // if
} // CalculateDuty

void loop() {
  static enum { ST_INITIAL,ST_WAIT,ST_HEAT1,ST_PREHEAT1,ST_PREHEAT2,ST_HEAT2,ST_REFLOW1,ST_REFLOW2,ST_COOL } state=ST_INITIAL; // ステートマシンの状態
  static int duty=0; // デューティー比
  static float temperature; // 測定点の温度
  static byte result; // 温度測定の際のエラー情報
  static int PwmCnt=0; // PWM用カウンタ
  static int GraphCnt=0; // グラフ更新用カウンタ
  static const char StartMessage[]="SW3 ｦ ｵｽﾄ ｽﾀｰﾄ";
  
  // 次のTickを待つ
  while(long(millis()-NextTick)<0);
  NextTick+=TIME_RESOLUTION;
  
  // ヒーター制御
  digitalWrite(HEATER_PIN,PwmCnt<duty ? HIGH : LOW);

  // 温度の測定
  if(PwmCnt==0) {
    int RetryCnt=0;
    do{
      result=MeasureTemperature(temperature);
    } while(result!=TEMP_NO_ERROR && RetryCnt++<RETRY_LIMIT);
    PrintTemperature(temperature,result);
  } // if
  
  // 時刻の表示
  if(state!=ST_INITIAL && state!=ST_WAIT) PrintTime();

  // グラフの更新  
  if(PwmCnt==1 && ++GraphCnt>=5) {
    GraphCnt=0;
    for(int i=1; i<GRAPH_WIDTH; i++) {
      TempHist[i-1]=TempHist[i];
    } // for i
    if(result!=TEMP_NO_ERROR || temperature<=GRAPH_LOW) {
      TempHist[GRAPH_WIDTH-1]=0;
    } else if(temperature>GRAPH_HIGH) {
      TempHist[GRAPH_WIDTH-1]=LCD_HEIGHT;
    } else {
      TempHist[GRAPH_WIDTH-1]=byte((temperature-GRAPH_LOW)/(GRAPH_HIGH-GRAPH_LOW)*LCD_HEIGHT);
    } // if
    for(int i=0; i<GRAPH_WIDTH; i++) {
      int x=LCD_WIDTH-GRAPH_WIDTH+i;
      if(TempHist[i]>0) {
        MGLCD.Line(x,LCD_HEIGHT-1,x,LCD_HEIGHT-TempHist[i],1);
      } // if
      if(TempHist[i]<LCD_HEIGHT) {
        MGLCD.Line(x,LCD_HEIGHT-TempHist[i]-1,x,0,0);
      } // if
    } // for i
  } // if
  
  // Serialに時刻と温度を送信
  if(state!=ST_INITIAL && state!=ST_WAIT && PwmCnt==2 && result==TEMP_NO_ERROR) {
    Serial.print(millis()-StartTime1);
    Serial.print(',');
    Serial.println(temperature);
  } // if
  
  // 強制中断処理  
  if((state==ST_HEAT1 || state==ST_PREHEAT1 || state==ST_PREHEAT2 || state==ST_HEAT2 || state==ST_REFLOW1 || state==ST_REFLOW2) && (MGLCD.ReadButton(3)==MGLCD_BUTTON_ON)) {
    PrintStatus("ｷｮｳｾｲ ﾁｭｳﾀﾞﾝ");
    tone(BUZZER_PIN,500,100);
    state=ST_COOL;
    StartTime2=millis();
    duty=0;    
  } // if

  // ステートの遷移
  switch(state) {
    case ST_INITIAL:
      PrintStatus(StartMessage);
      state=ST_WAIT;
      break;
    case ST_WAIT:
      if(result==TEMP_NO_ERROR && MGLCD.ReadButton(2)==MGLCD_BUTTON_ON) { // 温度が測定できていてかつSW3が押された
        PrintStatus("ｶﾈﾂ ｽﾃｰｼﾞ1");
        tone(BUZZER_PIN,1000,100);
        state=ST_HEAT1;
        StartTime1=StartTime2=millis();
        duty=PWM_RESOLUTION;
      } // if
      break;
    case ST_HEAT1:
      if(PwmCnt==0 && result==TEMP_NO_ERROR && temperature<TEMP_CHECK && millis()-StartTime2>=TIME_CHECK) { // 加熱不良を検出した
        PrintStatus("ｶﾈﾂﾌﾘｮｳ ｴﾗｰ");
        tone(BUZZER_PIN,500,2000);
        state=ST_COOL;
        StartTime2=millis();
        duty=0;
      } else if(PwmCnt==0 && result==TEMP_NO_ERROR && temperature>=TEMP_PREHEAT-WIDTH_PREHEAT/2.0) { // 予熱温度の下限に達した
        PrintStatus("ﾖﾈﾂ ｽﾃｰｼﾞ");
        state=ST_PREHEAT1;
        StartTime2=millis();
        duty=0;
      } // if
      break;  
    case ST_PREHEAT1:
      if(millis()-StartTime2>=TIME_PREHEAT1) {
        state=ST_PREHEAT2;
      } // if
      break;
    case ST_PREHEAT2:
      if(millis()-StartTime2>=TIME_PREHEAT2) {
        PrintStatus("ｶﾈﾂ ｽﾃｰｼﾞ2");        
        state=ST_HEAT2;
        StartTime2=millis();
        duty=PWM_RESOLUTION;
      } // if
      break;
    case ST_HEAT2:
      if(PwmCnt==0 && result==TEMP_NO_ERROR && temperature>=TEMP_REFLOW-WIDTH_REFLOW/2.0) { // リフロー温度の下限に達した
        PrintStatus("ﾘﾌﾛｰ ｽﾃｰｼﾞ");
        state=ST_REFLOW1;
        StartTime2=millis();
        duty=0;
      } // if
      break;
    case ST_REFLOW1:
      if(millis()-StartTime2>=TIME_REFLOW1) {
        state=ST_REFLOW2;
      } // if
      break;
    case ST_REFLOW2:
      if(millis()-StartTime2>=TIME_REFLOW2) {
         PrintStatus("ﾚｲｷｬｸ ｽﾃｰｼﾞ");
        tone(BUZZER_PIN,1000,2000);
         state=ST_COOL;
         StartTime2=millis();
         duty=0;
      } // if
      break;
    case ST_COOL:
      if(PwmCnt==0 && result==TEMP_NO_ERROR && temperature<=TEMP_COOL) { // 冷却終了温度に達した
        PrintStatus(StartMessage);
        tone(BUZZER_PIN,1000,2000);
        state=ST_WAIT;
      } // if
      break;
  } // switch(state)
  
  // 予熱時のデューティ比の計算
  if(PwmCnt==0 && state==ST_PREHEAT2) duty=CalculateDuty(temperature,TEMP_PREHEAT,WIDTH_PREHEAT);
  
  // リフロー時のデューティ比の計算
  if(PwmCnt==0 && state==ST_REFLOW2) duty=CalculateDuty(temperature,TEMP_REFLOW,WIDTH_REFLOW);
  
  // PWM用カウンタを進める
  if(++PwmCnt>=PWM_RESOLUTION) PwmCnt=0;
} // loop


